/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>

#include "rua.h"
#include "rua_dbus.h"
#include "rua_private.h"

#define RUA_INTERFACE "org.tizen.rua"
#define RUA_PATH "/org/tizen/rua"
#define RUA_SIGNAL_DATA_UPDATE "dataupdate"

struct cb_data {
	int callback_id;
	rua_history_update_cb callback;
	void *user_data;
};

static GDBusConnection *conn;
static guint s_id;

static gint atomic_callback_id;
static GHashTable *callback_hash_table;

static void __foreach_callback(gpointer key, gpointer value,
		gpointer user_data);
static void __signal_handler(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data);
static void __rua_dbus_init(void);
static void __rua_dbus_exit(void);

int rua_dbus_send_update_signal(update_type type)
{
	__rua_dbus_init();
	GError *err = NULL;

	if (g_dbus_connection_emit_signal(conn,
				NULL,
				RUA_PATH,
				RUA_INTERFACE,
				RUA_SIGNAL_DATA_UPDATE,
				g_variant_new("(i)", type),
				&err) == FALSE) {
		LOGE("g_dbus_connection_emit_signal() is failed. %s",
				err->message);
		g_clear_error(&err);
		__rua_dbus_exit();
		return -1;
	}

	if (g_dbus_connection_flush_sync(conn, NULL, &err) == FALSE) {
		LOGE("g_dbus_connection_flush_sync() is failed. %s",
				err->message);
		g_clear_error(&err);
		__rua_dbus_exit();
		return -1;
	}

	g_clear_error(&err);

	__rua_dbus_exit();

	return 0;
}

int rua_dbus_signal_subscribe(rua_history_update_cb callback,
		void *user_data, int *callback_id)
{
	struct cb_data *cd = NULL;

	if (s_id == 0) {
		__rua_dbus_init();

		s_id = g_dbus_connection_signal_subscribe(conn,
				NULL,
				RUA_INTERFACE,
				RUA_SIGNAL_DATA_UPDATE,
				RUA_PATH,
				NULL,
				G_DBUS_SIGNAL_FLAGS_NONE,
				__signal_handler,
				NULL,
				NULL);

		if (s_id == 0) {
			LOGE("g_dbus_connection_signal_subscribe() is failed.");
			__rua_dbus_exit();
			return -1;
		}

		if (!callback_hash_table) {
			callback_hash_table = g_hash_table_new_full(g_direct_hash, g_direct_equal,
					NULL, free);
			if (!callback_hash_table) {
				LOGE("out of memory : g_hash_table_new() is failed");
				__rua_dbus_exit();
				return -1;
			}
		}
	}

	cd = (struct cb_data *)malloc(sizeof(struct cb_data));
	if (!cd) {
		LOGE("out of memory : malloc() is failed");
		return -1;
	}

	g_atomic_int_inc(&atomic_callback_id);

	cd->callback = callback;
	cd->user_data = user_data;
	cd->callback_id = atomic_callback_id;

	g_hash_table_insert(callback_hash_table,
			GINT_TO_POINTER(cd->callback_id), (gpointer)cd);

	*callback_id = cd->callback_id;

	return 0;
}

int rua_dbus_signal_unsubscribe(int callback_id)
{
	gboolean result = FALSE;
	guint table_size = 0;

	result = g_hash_table_remove(callback_hash_table, GINT_TO_POINTER(callback_id));
	if (!result) {
		LOGE("g_hash_table_remove failed(%d is wrong)", callback_id);
		return -1;
	}

	table_size = g_hash_table_size(callback_hash_table);
	if (s_id && table_size == 0) {
		g_dbus_connection_signal_unsubscribe(conn, s_id);
		g_hash_table_destroy(callback_hash_table);
		__rua_dbus_exit();
		callback_hash_table = NULL;
		s_id = 0;
	}

	return 0;
}

static void __foreach_callback(gpointer key, gpointer value,
		gpointer user_data)
{
	char **table = NULL;
	int nrows = 0;
	int ncols = 0;
	int r = 0;

	struct cb_data *cd = (struct cb_data *)value;

	r = rua_history_load_db(&table, &nrows, &ncols);
	if (r == -1)
		return;

	if (cd->callback)
		cd->callback(table, nrows, ncols, cd->user_data);
}

static void __signal_handler(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{
	int update_type;

	LOGI("__signal_handler");
	if (g_strcmp0(signal_name, RUA_SIGNAL_DATA_UPDATE))
		return;

	g_variant_get(parameters, "(i)", &update_type);

	g_hash_table_foreach(callback_hash_table, __foreach_callback, NULL);
}

static void __rua_dbus_init(void)
{
	if (!conn) {
		GError *err = NULL;
		conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
		if (!conn) {
			LOGE("g_bus_get_sync() is failed. %s", err->message);
			g_error_free(err);
			return;
		}
	}
}

static void __rua_dbus_exit(void)
{
	if (conn) {
		g_object_unref(conn);
		conn = NULL;
	}
}
